/* Private data for Gr3270 widget */

#ifndef _Gr3270P_h
#define _Gr3270P_h

#include "Gr3270.h"
#include "X3270P.h"

/* New fields for the Gr3270 widget record */
typedef struct {
    Pixel gColor1;
    Pixel gColor2;
    Pixel gColor3;
    Pixel gColor4;
    Pixel gColor5;
    Pixel gColor6;
    Pixel gColor7;
    Pixel gColor8;
    Pixel gColor9;
    Pixel gColor10;
    Pixel gColor11;
    Pixel gColor12;
    Pixel gColor13;
    Pixel gColor14;
    Pixel gColor15;
    Pixel gColor16;
} Gr3270Part;

/* Full instance record declaration */
typedef struct _Gr3270Rec {
    CorePart	core;
    Win3270Part	win3270;
    X3270Part	x3270;
    Gr3270Part	gr3270;
} Gr3270Rec;

/* New fields for the Gr3270 widget class record */
typedef struct { int empty; } Gr3270ClassPart;

/* Full class record declaration */
typedef struct _Gr3270ClassRec {
    CoreClassPart    core_class;
    Win3270ClassPart win3270_class;
    X3270ClassPart   x3270_class;
    Gr3270ClassPart  gr3270_class;
} Gr3270ClassRec;

/* Class record variable */
extern Gr3270ClassRec gr3270ClassRec;

#endif /* _Gr3270P_h */
