/* Menu button widget definitions */

#ifndef _XtMenuBtn_h
#define _XtMenuBtn_h

#include "Command.h"

/* Class record pointer */
extern WidgetClass menuBtnWidgetClass;

/* Widget type definitions */
typedef struct _MenuBtnClassRec	*MenuBtnWidgetClass;
typedef struct _MenuBtnRec	*MenuBtnWidget;

#endif /* _XtMenuBtn_h */
