/* Tform widget definitions */

#ifndef _Tform_h
#define _Tform_h

#include <X11/Composite.h>

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 destroyCallback     Callback		Pointer		NULL
 height		     Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 sensitive	     Sensitive		Boolean		True
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
*/

/* Class record pointer */
extern WidgetClass tformWidgetClass;

/* Widget type definitions */
typedef struct _TformClassRec	*TformWidgetClass;
typedef struct _TformRec	*TformWidget;

#endif /* _Tform_h */
