/* Private data for Im3270 widget */

#ifndef _Im3270P_h
#define _Im3270P_h

#include "Im3270.h"
#include "Win3270P.h"

/* New fields for the Im3270 widget record */
typedef struct {
    XtTranslations *imTranslations;
} Im3270Part;

/* Full instance record declaration */
typedef struct _Im3270Rec {
    CorePart	core;
    Win3270Part	win3270;
    Im3270Part	im3270;
} Im3270Rec;

/* New fields for the Im3270 widget class record */
typedef struct { int empty; } Im3270ClassPart;

/* Full class record declaration */
typedef struct _Im3270ClassRec {
    CoreClassPart    core_class;
    Win3270ClassPart win3270_class;
    Im3270ClassPart  im3270_class;
} Im3270ClassRec;

/* Class record variable */
extern Im3270ClassRec im3270ClassRec;

#endif /* _Im3270P_h */
