/* Gr3270 widget definitions */

#ifndef _Gr3270_h
#define _Gr3270_h

#include "X3270.h"

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 gColor1	     GColor1		Pixel		Blue
 gColor2	     GColor2		Pixel		Red
 gColor3	     GColor3		Pixel		Magenta
 gColor4	     GColor4		Pixel		Green
 gColor5	     GColor5		Pixel		Turquoise
 gColor6	     GColor6		Pixel		Yellow
 gColor7	     GColor7		Pixel		White
 gColor8	     GColor8		Pixel		Black
 gColor9	     GColor9		Pixel		DarkSlateBlue
 gColor10	     GColor10		Pixel		Orange
 gColor11	     GColor11		Pixel		Plum
 gColor12	     GColor12		Pixel		DarkGreen
 gColor13	     GColor13		Pixel		DarkTurquoise
 gColor14	     GColor14		Pixel		Wheat
 gColor15	     GColor15		Pixel		Gray
 gColor16	     GColor16		Pixel		Brown
*/

/* New resource definitions */
#define XtNgColor1	"gColor1"
#define XtNgColor2	"gColor2"
#define XtNgColor3	"gColor3"
#define XtNgColor4	"gColor4"
#define XtNgColor5	"gColor5"
#define XtNgColor6	"gColor6"
#define XtNgColor7	"gColor7"
#define XtNgColor8	"gColor8"
#define XtNgColor9	"gColor9"
#define XtNgColor10	"gColor10"
#define XtNgColor11	"gColor11"
#define XtNgColor12	"gColor12"
#define XtNgColor13	"gColor13"
#define XtNgColor14	"gColor14"
#define XtNgColor15	"gColor15"
#define XtNgColor16	"gColor16"

#define XtCGColor1	"GColor1"
#define XtCGColor2	"GColor2"
#define XtCGColor3	"GColor3"
#define XtCGColor4	"GColor4"
#define XtCGColor5	"GColor5"
#define XtCGColor6	"GColor6"
#define XtCGColor7	"GColor7"
#define XtCGColor8	"GColor8"
#define XtCGColor9	"GColor9"
#define XtCGColor10	"GColor10"
#define XtCGColor11	"GColor11"
#define XtCGColor12	"GColor12"
#define XtCGColor13	"GColor13"
#define XtCGColor14	"GColor14"
#define XtCGColor15	"GColor15"
#define XtCGColor16	"GColor16"

/* Class record pointer */
extern WidgetClass gr3270WidgetClass;

/* Widget type definitions */
typedef struct _Gr3270ClassRec	*Gr3270WidgetClass;
typedef struct _Gr3270Rec	*Gr3270Widget;

#endif /* _Gr3270_h */
