/* Ga3270 widget definitions */

#ifndef _Ga3270_h
#define _Ga3270_h

#include "Win3270.h"

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 destroyCallback     Callback		Pointer		NULL
 height		     Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 sensitive	     Sensitive		Boolean		True
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
 foreground	     Foreground		Pixel		XtDefaultForeground
 backColor	     BackColor		Pixel		XtDefaultForeground
 foreColor	     ForeColor		Pixel		XtDefaultBackground
 mouseColor	     MouseColor		Pixel		foreground or foreColor
 cursor		     Cursor		Cursor		XC_left_ptr
 speckey	     Boolean		Boolean		False
 keyfile 	     File		String		NULL
 keyProc 	     Callback		Pointer		NULL
 resizeProc 	     Callback		Pointer		NULL
 drawColor	     DrawColor		Pixel		White
 drawColor1-15	     DrawColor		Pixel		White
 refreshColor	     RefreshColor	Pixel		Cyan
 gaCursor	     GaCursor		Cursor		XC_crosshair
 gaTranslations      Translations       String          NULL
*/

/* New resource definitions */
#define XtNdrawColor	"drawColor"
#define XtNdrawColor1   "drawColor1"
#define XtNdrawColor2   "drawColor2"
#define XtNdrawColor3   "drawColor3"
#define XtNdrawColor4   "drawColor4"
#define XtNdrawColor5   "drawColor5"
#define XtNdrawColor6   "drawColor6"
#define XtNdrawColor7   "drawColor7"
#define XtNdrawColor8   "drawColor8"
#define XtNdrawColor9   "drawColor9"
#define XtNdrawColor10  "drawColor10"
#define XtNdrawColor11  "drawColor11"
#define XtNdrawColor12  "drawColor12"
#define XtNdrawColor13  "drawColor13"
#define XtNdrawColor14  "drawColor14"
#define XtNdrawColor15  "drawColor15"
#define XtNrefreshColor	"refreshColor"
#define XtNgaCursor	"gaCursor"
#define XtNgaTranslations "gaTranslations"

#define XtCDrawColor	"DrawColor"
#define XtCRefreshColor	"RefreshColor"
#define XtCGaCursor	"GaCursor"

/* Drawing types for XtGa3270DrawLine */
enum draw_types { dt_draw, dt_refresh, dt_refclear };

/* Class record pointer */
extern WidgetClass ga3270WidgetClass;

/* Widget type definitions */
typedef struct _Ga3270ClassRec	*Ga3270WidgetClass;
typedef struct _Ga3270Rec	*Ga3270Widget;

#endif /* _Ga3270_h */
