/* Private data for menu button widget */

#ifndef _XtMenuBtnP_h
#define _XtMenuBtnP_h

#include "MenuBtn.h"
#include "CommandP.h"

/* New fields for the MenuBtn widget record */
typedef struct { int empty; } MenuBtnPart;

/* Full instance record declaration */
typedef struct _MenuBtnRec {
    CorePart         core;
    SimplePart	     simple;
    LabelPart	     label;
    CommandPart	     command;
    MenuBtnPart      menuBtn;
} MenuBtnRec;

/* New fields for the MenuBtn widget class record */
typedef struct { int empty; } MenuBtnClassPart;

/* Full class record declaration */
typedef struct _MenuBtnClassRec {
    CoreClassPart	core_class;
    SimpleClassPart	simple_class;
    LabelClassPart	label_class;
    CommandClassPart	command_class;
    MenuBtnClassPart    menuBtn_class;
} MenuBtnClassRec;

/* Class record variable */
extern MenuBtnClassRec menuBtnClassRec;

#endif /* _XtMenuBtnP_h */


