/* 3270 screen title widget definitions */

#ifndef _XtTitle3270_h
#define _XtTitle3270_h

#include <X11/Core.h>

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	0
 destroyCallback     Callback		Pointer		NULL
 height		     Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 sensitive	     Sensitive		Boolean		True
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
 aplkeyb             Aplkeyb            Boolean         False
 cursor		     Cursor		Cursor		XC_top_left_arrow
 cursorColor	     CursorColor	Pixel		XtDefaultForeground
 foreground	     Foreground		Pixel		XtDefaultForeground
 font		     Font		FontStruct	XtDefaultFont
 inserting	     Inserting		Boolean		False
 internalWidth	     Width		Dimension	4
 internalHeight	     Height		Dimension	2
 system		     System		Boolean		True
 title		     Title		String		NULL
*/

/* New resource definitions */
#define XtNaplkeyb      "aplkeyb"
#define XtNcursor	"cursor"
#define XtNcursorColor	"cursorColor"
#define XtNinternalHeight "internalHeight"
#define XtNinternalWidth "internalWidth"
#define XtNinserting	"inserting"
#define XtNsystem	"system"
#define XtNtitle	"title"

#define XtCAplkeyb      "Aplkeyb"
#define XtCCursorColor	"CursorColor"
#define XtCInserting	"Inserting"
#define XtCSystem	"System"
#define XtCTitle	"Title"

/* Class record pointer */
extern WidgetClass title3270WidgetClass;

/* Widget type definitions */
typedef struct _Title3270ClassRec  *Title3270WidgetClass;
typedef struct _Title3270Rec	   *Title3270Widget;

#endif /* _XtTitle3270_h */
