/* Im3270 widget definitions */

#ifndef _Im3270_h
#define _Im3270_h

#include "Win3270.h"

/* Resources:

 Name		     Class		RepType		Default Value
 ----		     -----		-------		-------------
 background	     Background		Pixel		XtDefaultBackground
 border		     BorderColor	Pixel		XtDefaultForeground
 borderWidth	     BorderWidth	Dimension	1
 destroyCallback     Callback		Pointer		NULL
 height		     Height		Dimension	0
 mappedWhenManaged   MappedWhenManaged	Boolean		True
 sensitive	     Sensitive		Boolean		True
 width		     Width		Dimension	0
 x		     Position		Position	0
 y		     Position		Position	0
 foreground	     Foreground		Pixel		XtDefaultForeground
 backColor	     BackColor		Pixel		XtDefaultForeground
 foreColor	     ForeColor		Pixel		XtDefaultBackground
 mouseColor	     MouseColor		Pixel		foreground or foreColor
 cursor		     Cursor		Cursor		XC_left_ptr
 speckey	     Boolean		Boolean		False
 keyfile 	     File		String		NULL
 keyProc 	     Callback		Pointer		NULL
 resizeProc 	     Callback		Pointer		NULL
 imTranslations      Translations       String          NULL
 colormap	     Colormap	        Colormap        NULL
*/

/* New resource definitions */
#define XtNimTranslations "imTranslations"

/* Class record pointer */
extern WidgetClass im3270WidgetClass;

/* Widget type definitions */
typedef struct _Im3270ClassRec	*Im3270WidgetClass;
typedef struct _Im3270Rec	*Im3270Widget;

#endif /* _Im3270_h */
