/* Private data for Tform widget */

#ifndef _TformP_h
#define _TformP_h

#include "Tform.h"
#include <X11/CompositeP.h>

/* New fields for the Tform widget record */
typedef struct {
    int j;			/* ??TEMPORARY */
} TformPart;

/* Full instance record declaration */
typedef struct _TformRec {
    CorePart	  core;
    CompositePart composite;
    TformPart	  tform;
} TformRec;

/* New fields for the Tform widget class record */
typedef struct { int empty; } TformClassPart;

/* Full class record declaration */
typedef struct _TformClassRec {
    CoreClassPart      core_class;
    CompositeClassPart composite_class;
    TformClassPart     tform_class;
} TformClassRec;

/* Class record variable */
extern TformClassRec tformClassRec;

#endif /* _TformP_h */
